/**************************************************************************
 * CERN, GSI Darmstadt                                                    *
 * All rights reserved.                                                   *
 *                                                                        *
 * Permission to use, copy, modify and distribute this software and its   *
 * documentation strictly for non-commercial purposes is hereby granted   *
 * without fee, provided that the above copyright notice appears in all   *
 * copies and that both the copyright notice and this permission notice   *
 * appear in the supporting documentation. The authors make no claims     *
 * about the suitability of this software for any purpose. It is          *
 * provided "as is" without express or implied warranty.                  *
 **************************************************************************/
#include <iostream>     // for output messages
#include <limits>       // for data limits
#include <windows.h>    // for DWORD declarator
#include <mmsystem.h>   // for timeGetTime() function
#include "fit_util.h"   // a set of constants

#undef max      // undefine the max and min set up within windows
#undef min      // so that standard max and min can be used

typedef float ftype;  // set ftype to be single precision data

using namespace std;

// ------------------------- Prototypes
void readInput( int * nT, int * nS);
void fitTracks( ftype (*T)[6], ftype (*C)[15], int nT, int nS );

// ******************************************************************
int main(int /*argc*/, char* /*argv*/[])
{
    int i, nT, nS;   // loop counter, number of tracks & stations
	DWORD StartTime; // Start time
	double Timing, Elapsed;  // Timing values
// ------------------------------------------------------------------
	try
	{
//                           Read data from magnetic and track files
//                           Also sets up dynamic arrays whose addresses
//                           are loaded into the class pointer data
        readInput( &nT, &nS );
//                           Set up pointer arrays for parameters
//                           For each track set up array for State matrix
//                             T1[6] = x, y, tx, ty, qp, z
        ftype (*T1)[6]  = new ftype[nT][6];
//                            For each track set up array for Covariance matrix C1[15]
		ftype (*C1)[15] = new ftype[nT][15];

// ------------------------- Run Serial version
//                           Set timimg to maximum value for a double
        Timing = std::numeric_limits<double>::max();
//                           Repeat NUM_RUNS times
		for(i=0; i<NUM_RUNS;i++)
		{
//                           set start time
			StartTime = timeGetTime();
//                           call main serial track fitting function
			fitTracks( T1, C1, nT, nS );
//                           determine elapsed time
			Elapsed = (int)(timeGetTime() - StartTime);
//                           get minimum time so far
            Timing = std::min(Timing, Elapsed);
        }
//                           Report serial timings
        printf( "----------------------------------------\n" );
        printf( "  Version    Minimum Time(s)\n" );
        printf( "  Serial     %12.6f\n", 0.001 * Timing );
        printf( "----------------------------------------\n" );

//                           release space
        delete []T1;
        delete []C1;
//                           Finish message
        printf( "Test finished\n");
    }                        // end of try block
//                           Service Exception catches
	catch (const std::exception& e)
	{
        std::cerr << "Error: Caught unexpected std::exception." << std::endl;
        std::cerr << "\t"<< e.what() << std::endl;
        return EXIT_FAILURE;
    }
	catch (...)
	{
        std::cerr << "Error: Caught unexpected exception" << std::endl;
        std::cerr << "\t" << "<unknown-exception-type>" << std::endl;
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}

